/*
** Module   :CLIPBRD.CPP
** Abstract :Clipboard-related methods of class Buffer
**
** Copyright (C) Sergey I. Yevtushenko
**
** Log: Wed  05/03/1997     Updated to V0.5
**      Sat  12/04/1997   	Updated to V0.8
**      Wed  12/11/1997   	Updated to V0.9, mostly rewritten
*/

#include <buffer.h>
#include <stdlib.h>
#include <string.h>
#include <version.h>

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

Buffer* Buffer::copy()
{
    if(!mark_state)
        return 0;

    int mark_beg_row = min(old_abs_row, abs_row());
    int mark_end_row = max(old_abs_row, abs_row());

    Buffer* clip = new Buffer;
    if(col_block || mark_beg_row == mark_end_row)
    {
        int mark_col_start = min(old_abs_col, abs_col());
        int mark_col_end   = max(old_abs_col, abs_col());

        if(col_block)
        {
            clip->set_column_block(1);
            clip->num_cols = mark_col_end - mark_col_start;
        }

        char *tmp = new char [(mark_col_end - mark_col_start + 1)];

        for(int i = mark_beg_row; i <= mark_end_row; i++)
        {
            line(i)->get_print(mark_col_start,
                               tmp,
                               mark_col_end - mark_col_start);
            tmp[mark_col_end - mark_col_start] = 0;
            PLine ln = new Line;
            ln->set(tmp);
            ln->xlat(cp_out);
            clip->add_line(ln);
        }
    }
    else
    {
        for(int i = mark_beg_row; i <= mark_end_row; i++)
        {
            PLine ln = new Line;
            if(i == mark_beg_row)
            {
                int col_start = (mark_beg_row == old_abs_row)
                            	? old_abs_col : abs_col();
                int j = 0;

                for(int chr = line(i)->char_at(col_start);
                    chr;
                    chr = line(i)->char_at(++col_start), j++)

                    ln->ins_char(chr, j);

                ln->xlat(cp_out);
                clip->add_line(ln);
                continue;
            }
            if(i == mark_end_row)
            {
                int col_end = (mark_end_row == old_abs_row)
                              ? old_abs_col : abs_col();

                int j = 0;

                for(int chr = line(i)->char_at(j);
                    chr && j < col_end;
                    chr = line(i)->char_at(++j))

                    ln->ins_char(chr, j);

                ln->xlat(cp_out);
                clip->add_line(ln);
                continue;
            }
            //Default action
            ln->set(line(i)->str);
            ln->xlat(cp_out);
            clip->add_line(ln);
        }
    }
    clip->to_pm();
    return clip;
}

Buffer* Buffer::cut(Rect& rect)
{
    if(!mark_state)
        return 0;
    Buffer* clip = copy();
    clear(rect);
    changed = 1;
    return clip;
}

void Buffer::clear(Rect& rect)
{
    if(!mark_state)
        return;

    changed = 1;
    int mark_beg_row = min(old_abs_row, abs_row());
    int mark_end_row = max(old_abs_row, abs_row());


    if(col_block || mark_beg_row == mark_end_row)
    {
        int mark_col_start = min(old_abs_col, abs_col());
        int mark_col_end   = max(old_abs_col, abs_col());

        for(int i = mark_beg_row; i <= mark_end_row; i++)
        {
            track(opRestoreLine,(void *)line(i),(void *)i);
            for(int j = mark_col_start; j < mark_col_end; j++)
            {
                line(i)->del_char(mark_col_start);
            }
        }

        if(abs_col() != mark_col_start)
            goto_col(rect, mark_col_start);
    }
    else
    {
        int i;
        for(i = mark_beg_row; i <= mark_end_row; i++)
        {
            int col_start = 0;
            int col_end = line(i)->len();

            if(i != mark_beg_row && i != mark_end_row)
                continue;
            if(i == mark_beg_row)
                col_start = (mark_beg_row == old_abs_row)
                    ? old_abs_col : abs_col();

            if(i == mark_end_row)
                col_end = (mark_end_row == old_abs_row)
                	? old_abs_col : abs_col();

            track(opRestoreLine,(void *)line(i),(void *)i);

            for(int j = col_start; j < col_end; j++)
            {
                line(i)->del_char(col_start);
            }
        }

        i = mark_end_row - (mark_beg_row + 1);

        if(i > 0)
        {
            Buffer *tmp_buf = new Buffer(i);
            remove_items(tmp_buf,(mark_beg_row + 1), i);
            track(opInsBlock,
                  (void *) tmp_buf,
                  (void *) (mark_beg_row + 1));
        }

        goto_line(rect, mark_beg_row);

        line_end(rect);
        del_char(rect);
    }
    unmark();

    int in_comment = line(mark_beg_row)->in_comment;
    int in_quote   = line(mark_beg_row)->in_quote;
    fill_rest_conditions(mark_beg_row, in_comment, in_quote);
}

void Buffer::paste(Rect& rect, Buffer* clip)
{
    if(!clip)
        return;

    changed = 1;
    int recalc_row = abs_row()-1;

    if(recalc_row < 0)
        recalc_row = 0;

    int i,j;

    clip->from_pm();

    if(clip->col_block || clip->Count() == 1)
    {
        for(i = 0; i < clip->Count(); i++)
        {
            PLine ln = line(abs_row() + i);

            track(opRestoreLine,(void *)ln,(void *)(abs_row() + i));

            for(j = 0; j < clip->line(i)->len(); j++)
                ln->ins_char(chr_in(clip->line(i)->char_at(j)), abs_col()+j);

            if((i < clip->Count() - 1) && (abs_row() + i) == (Count() - 1))
            {
                track(opDelLine,(void *)(Count()));
                Add(new Line);
            }
        }
    }
    else
    {
        PLine ln;

        for(i = 1; i < clip->Count() - 1; i++)
        {
            track(opDelLine,(void *)(abs_row() + i));
            ln = new Line(clip->line(i));
            ln->xlat(cp_in);

            At(ln, abs_row() + i);
        }

        ln = line(abs_row());
        track(opRestoreLine,(void *)ln,(void *)(abs_row()));

        PLine ln0 = new Line;
        char chr;

        do
        {
            chr = chr_out((char)ln->del_char(abs_col()));
            ln0->ins_char(chr_in(chr),ln0->len());
        }
        while(chr);

        ins_line(ln0, abs_row()+clip->Count()-1);
        track(opDelLine,(void *)(abs_row()+clip->Count()-1));

        for(j = 0; j < clip->line(0)->len(); j++)
            ln->ins_char(chr_in(clip->line(0)->char_at(j)), abs_col()+j);

        for(j = 0; j < clip->line(clip->Count()-1)->len(); j++)
            ln0->ins_char(chr_in(clip->line(clip->Count()-1)->char_at(j)), j);

        goto_line(rect, abs_row() + clip->Count()-1);
        goto_col(rect, clip->line(clip->Count()-1)->len());
    }
    int in_comment = line(recalc_row)->in_comment;
    int in_quote   = line(recalc_row)->in_quote;
    fill_rest_conditions(recalc_row, in_comment, in_quote);
}

void Buffer::paste_over(Rect&,Buffer*)
{
    //Still unimplemented
    changed = 1;
}
