/*
** Module   :LINE.CPP
** Abstract :Editor Line methods
**
** Copyright (C) Sergey I. Yevtushenko
**
** Log: Wed  05/03/1997   	Updated to V0.5
**      Sun  09/11/1997   	Updated to V0.9
*/

#include <common.h>
#include <line.h>
#include <stdlib.h>
#include <string.h>
#include <version.h>

Line::Line()
{
    str = new char[CHUNK_SIZE];
    if(str)
    {
        str[0] = 0;
        buf_len = CHUNK_SIZE;
    }
    else
        buf_len = 0;
    in_comment = in_quote = 0;
    known_len = 0;
}

Line::Line(PLine master)
{
    if(master->buf_len)
    {
        str = new char[master->buf_len];
        memcpy(str, master->str, master->buf_len);

        buf_len    = master->buf_len;
    }
    else
    {
        str        = master->str;
        buf_len    = 0;
    }

    in_comment = in_quote = 0;
    known_len  = master->known_len;
}

Line::Line(char *new_str)
{
    str = new_str;
    buf_len = 0;
    in_comment = in_quote = 0;
    known_len = 0;
}

Line::~Line()
{
    if(str && buf_len)
        delete str;
}

void Line::expand_by(int len)
{
    char *tmp;
    len = (len / CHUNK_SIZE + 1) * CHUNK_SIZE;
    tmp = new char [buf_len + len];
    if(tmp)
    {
        if(str)
        {
            memcpy(tmp, str, buf_len);
            delete str;
        }
        buf_len += len + ((str) ? 0:1);
        str = tmp;
    }
    /* Handle Error (NOT_ENOUGH_MEMORY)
    */

    known_len = 0;
}
void Line::set(char* str_new)
{
    int len = strlen(str_new)+1;
    if((len+1) > buf_len)
        expand_by(len - buf_len + 1);
    if(str)
        memcpy(str, str_new, len + 1);

    known_len = 0;
}

char* Line::get_print(int start, char *buff, int width)
{
    char *tmp = new char [(len() + 1)];

    if(start < len())
    {
        build_print(tmp);

        memmove(buff, &tmp[start], width < len() ? width:len());

        if((start + width) > len())
        {
            int from = len() - start;
            int cnt  = width - from;
            memset(&buff[from], ' ', cnt);
        }
    }
    else
        memset(buff, ' ', width);

    buff[width] = 0;
    delete tmp;

    return buff;
}

void Line::touch()
{
    int slen;
    char *tmp;

    if(!buf_len)
    {
        slen = len();
        slen = (slen / CHUNK_SIZE + 1) * CHUNK_SIZE;
        tmp = new char[slen];
        buf_len = slen;
        strcpy(tmp, str);
        str = tmp;
    }

    tmp = strchr(str, '\t');
    if(tmp)
    {
        int i = 1;
        while(tmp)
        {
            tmp = strchr(tmp + 1, '\t');
            i++;
        }
        expand_by(i * TAB_WIDTH);

        tmp = new char [buf_len];
        build_print(tmp);
        tmp[len()] = 0;
        set(tmp);
        delete tmp;
    }

}

int Line::ins_char(int chr, int ins_pos)
{
    int slen;
    if(ins_pos < 0)
        return 1;

    touch();

    slen = strlen(str);
    if(slen + 1 >= buf_len)
        expand_by(CHUNK_SIZE);
    if(ins_pos > slen)
    {
        expand_by(ins_pos - slen);
        memset(&str[slen], ' ', ins_pos - slen);
        str[ins_pos] = 0;
        slen = ins_pos;
    }
    memmove(&str[ins_pos + 1], &str[ins_pos], slen - ins_pos + 1);
    str[ins_pos] = (char)chr;
    known_len = 0;

    return (chr == '\t') ? ((ins_pos / TAB_WIDTH + 1) * TAB_WIDTH - ins_pos) : 1;
}

int Line::del_char(int del_pos)
{
    int slen;
    int chr = 0;

    if(del_pos < 0 || !strlen(str))
        return 0;

    touch();

    slen = strlen(str);
    if(del_pos >= slen)
        return 0;
    chr = str[del_pos];
    memcpy(&str[del_pos], &str[del_pos+1], slen - del_pos + 1);
    known_len = 0;

    return chr;
}

int Line::len()
{
    if(known_len)
        return known_len;

    int pos = 0;
    char *str2 = str;

    if(str2) /* Buffer is not empty */
    {
        for(pos = 0; *str2; str2++)
        {
            if(*str2 == '\t')
                pos = ((pos / TAB_WIDTH + 1) * TAB_WIDTH);
            else
                pos++;
        }
    }

    return (known_len = pos);
}

int Line::char_at(int from_pos)
{
    if(from_pos >= len())
        return 0;
/*
    char *tmp = new char[len() + 1];

    build_print(tmp);

    int res = tmp[pos];
    delete tmp;
*/
    char *str2 = str;
    int pos = 0;

    if(str2)
    {
        for(pos = 0; *str2; str2++)
        {
            if(*str2 == '\t')
            {
                int fill = pos;
                pos = ((pos / TAB_WIDTH + 1) * TAB_WIDTH);

                if(fill <= from_pos && pos > from_pos)
                    return ' ';
            }
            else
            {
                if(pos == from_pos)
                    return *str2;

                pos++;
            }
        }
    }

    return 0;
}

void Line::build_print(char *buff)
{
    char *str2 = str;
    int pos = 0;

    if(str2)
    {
        for(pos = 0; *str2; str2++)
        {
            if(*str2 == '\t')
            {
                int fill = pos;
                pos = ((pos / TAB_WIDTH + 1) * TAB_WIDTH);

                while(fill++ < pos)
                    *buff++ = ' ';
            }
            else
            {
                *buff++ = *str2;
                pos++;
            }
        }
    }
    *buff = 0;
}

void Line::xlat(char *cvt_tbl)
{
    int slen;

    touch();

    slen = strlen(str);

    for(int x = 0; x < slen; x++)
        str[x] = cvt_tbl[str[x]];
}
